@@ -1,7 +1,6 @@
 /******************************************************************************
- * $Id$
  *
- * Project:  MapServer
+ * Project:  MapCache
  * Purpose:  MapCache tile caching support file: Berkeley DB cache backend
  * Author:   Thomas Bonfort and the MapServer team.
  *
@@ -27,10 +26,8 @@
  * DEALINGS IN THE SOFTWARE.
  *****************************************************************************/
 
-#include "mapcache-config.h"
-#ifdef USE_BDB
-
 #include "mapcache.h"
+#ifdef USE_BDB
 #include <apr_strings.h>
 #include <apr_reslist.h>
 #include <apr_file_info.h>
@@ -51,6 +48,13 @@
 #define PAGESIZE 64*1024
 #define CACHESIZE 1024*1024
 
+typedef struct mapcache_cache_bdb mapcache_cache_bdb;
+struct mapcache_cache_bdb {
+  mapcache_cache cache;
+  char *basedir;
+  char *key_template;
+};
+
 struct bdb_env {
   DB* db;
   DB_ENV *env;
@@ -406,11 +410,11 @@ mapcache_cache* mapcache_cache_bdb_create(mapcache_context *ctx)
   }
   cache->cache.metadata = apr_table_make(ctx->pool,3);
   cache->cache.type = MAPCACHE_CACHE_BDB;
-  cache->cache.tile_delete = _mapcache_cache_bdb_delete;
-  cache->cache.tile_get = _mapcache_cache_bdb_get;
-  cache->cache.tile_exists = _mapcache_cache_bdb_has_tile;
-  cache->cache.tile_set = _mapcache_cache_bdb_set;
-  cache->cache.tile_multi_set = _mapcache_cache_bdb_multiset;
+  cache->cache._tile_delete = _mapcache_cache_bdb_delete;
+  cache->cache._tile_get = _mapcache_cache_bdb_get;
+  cache->cache._tile_exists = _mapcache_cache_bdb_has_tile;
+  cache->cache._tile_set = _mapcache_cache_bdb_set;
+  cache->cache._tile_multi_set = _mapcache_cache_bdb_multiset;
   cache->cache.configuration_post_config = _mapcache_cache_bdb_configuration_post_config;
   cache->cache.configuration_parse_xml = _mapcache_cache_bdb_configuration_parse_xml;
   cache->basedir = NULL;
@@ -418,6 +422,11 @@ mapcache_cache* mapcache_cache_bdb_create(mapcache_context *ctx)
   return (mapcache_cache*)cache;
 }
 
+#else
+mapcache_cache* mapcache_cache_bdb_create(mapcache_context *ctx) {
+  ctx->set_error(ctx,400,"BERKELEYDB support not compiled in this version");
+  return NULL;
+}
 #endif
 
 /* vim: ts=2 sts=2 et sw=2
