@@ -47,47 +47,6 @@ void parseMetadata(mapcache_context *ctx, ezxml_t node, apr_table_t *metadata)
   }
 }
 
-void parseTimeDimension(mapcache_context *ctx, ezxml_t node, mapcache_tileset *tileset) {
-  const char *attr = NULL;
-  if(tileset->timedimension) {
-    ctx->set_error(ctx,400,"tileset \"%s\" can only have a single <timedimension>", tileset->name);
-    return;
-  }
-  attr = ezxml_attr(node,"type");
-  if(attr && *attr) {
-    if(!strcmp(attr,"sqlite")) {
-#ifdef USE_SQLITE
-      tileset->timedimension = mapcache_timedimension_sqlite_create(ctx->pool);
-#else
-      ctx->set_error(ctx,400, "failed to add sqlite timedimension: Sqlite support is not available on this build");
-      return;
-#endif
-    } else {
-      ctx->set_error(ctx,400,"unknown \"type\" attribute \"%s\" for %s's <timedimension>. Expecting one of (sqlite)",attr,tileset->name);
-      return;
-    }
-    
-  } else {
-    ctx->set_error(ctx,400,"missing \"type\" attribute for %s's <timedimension>",tileset->name);
-    return;
-  }
-  attr = ezxml_attr(node,"name");
-  if(attr && *attr) {
-    tileset->timedimension->key = apr_pstrdup(ctx->pool,attr);
-  } else {
-    tileset->timedimension->key = apr_pstrdup(ctx->pool,"TIME");
-  }
-  
-  attr = ezxml_attr(node,"default");
-  if(attr && *attr) {
-    tileset->timedimension->default_value = apr_pstrdup(ctx->pool,attr);
-  } else {
-    ctx->set_error(ctx,400,"no \"default\" attribute for <timedimension> %s",tileset->timedimension->key);
-    return;
-  }
-  tileset->timedimension->configuration_parse_xml(ctx,tileset->timedimension,node);
-}
-
 void parseDimensions(mapcache_context *ctx, ezxml_t node, mapcache_tileset *tileset)
 {
   ezxml_t dimension_node;
@@ -96,7 +55,6 @@ void parseDimensions(mapcache_context *ctx, ezxml_t node, mapcache_tileset *tile
     char *name = (char*)ezxml_attr(dimension_node,"name");
     char *type = (char*)ezxml_attr(dimension_node,"type");
     char *unit = (char*)ezxml_attr(dimension_node,"unit");
-    char *skip_validation = (char*)ezxml_attr(dimension_node,"skip_validation");
     char *default_value = (char*)ezxml_attr(dimension_node,"default");
 
     mapcache_dimension *dimension = NULL;
@@ -108,17 +66,13 @@ void parseDimensions(mapcache_context *ctx, ezxml_t node, mapcache_tileset *tile
 
     if(type && *type) {
       if(!strcmp(type,"values")) {
-        dimension = mapcache_dimension_values_create(ctx->pool);
+        dimension = mapcache_dimension_values_create(ctx,ctx->pool);
       } else if(!strcmp(type,"regex")) {
-        dimension = mapcache_dimension_regex_create(ctx->pool);
-      } else if(!strcmp(type,"intervals")) {
-        dimension = mapcache_dimension_intervals_create(ctx->pool);
+        dimension = mapcache_dimension_regex_create(ctx,ctx->pool);
       } else if(!strcmp(type,"sqlite")) {
-        dimension = mapcache_dimension_sqlite_create(ctx->pool);
+        dimension = mapcache_dimension_sqlite_create(ctx,ctx->pool);
       } else if(!strcmp(type,"time")) {
-        ctx->set_error(ctx,501,"time dimension type not implemented yet");
-        return;
-        dimension = mapcache_dimension_time_create(ctx->pool);
+        dimension = mapcache_dimension_time_create(ctx,ctx->pool);
       } else {
         ctx->set_error(ctx,400,"unknown dimension type \"%s\"",type);
         return;
@@ -127,6 +81,7 @@ void parseDimensions(mapcache_context *ctx, ezxml_t node, mapcache_tileset *tile
       ctx->set_error(ctx,400, "mandatory attribute \"type\" not found in <dimensions>");
       return;
     }
+    GC_CHECK_ERROR(ctx);
 
     dimension->name = apr_pstrdup(ctx->pool,name);
 
@@ -134,16 +89,13 @@ void parseDimensions(mapcache_context *ctx, ezxml_t node, mapcache_tileset *tile
       dimension->unit = apr_pstrdup(ctx->pool,unit);
     }
     
-    if(skip_validation && !strcmp(skip_validation,"true")) {
-      dimension->skip_validation = MAPCACHE_TRUE;
-    }
-
     if(default_value && *default_value) {
       dimension->default_value = apr_pstrdup(ctx->pool,default_value);
     } else {
       ctx->set_error(ctx,400,"dimension \"%s\" has no \"default\" attribute",dimension->name);
       return;
     }
+    
 
     dimension->configuration_parse_xml(ctx,dimension,dimension_node);
     GC_CHECK_ERROR(ctx);
@@ -154,7 +106,51 @@ void parseDimensions(mapcache_context *ctx, ezxml_t node, mapcache_tileset *tile
     ctx->set_error(ctx, 400, "<dimensions> for tileset \"%s\" has no dimensions defined (expecting <dimension> children)",tileset->name);
     return;
   }
+  
   tileset->dimensions = dimensions;
+  dimension_node = ezxml_child(node,"store_assemblies");
+  if(dimension_node && dimension_node->txt) {
+    if(!strcmp(dimension_node->txt,"false")) {
+      tileset->store_dimension_assemblies = 0;
+    } else if(strcmp(dimension_node->txt,"true")) {
+      ctx->set_error(ctx,400,"failed to parse <store_assemblies> (%s), expecting \"true\" or \"false\"",dimension_node->txt);
+      return;
+    }
+  }
+  
+  dimension_node = ezxml_child(node,"assembly_type");
+  if(dimension_node) {
+    if(!strcmp(dimension_node->txt,"stack")) {
+      tileset->dimension_assembly_type = MAPCACHE_DIMENSION_ASSEMBLY_STACK;
+    } else if(!strcmp(dimension_node->txt,"animate")) {
+      tileset->dimension_assembly_type = MAPCACHE_DIMENSION_ASSEMBLY_ANIMATE;
+      ctx->set_error(ctx,400,"animate dimension assembly mode not implemented");
+      return;
+    } else if(strcmp(dimension_node->txt,"none")) {
+      ctx->set_error(ctx,400,"unknown dimension assembly mode (%s). Can be one of \"stack\" or \"none\"",dimension_node->txt);
+      return;
+    } else {
+      tileset->dimension_assembly_type = MAPCACHE_DIMENSION_ASSEMBLY_NONE;
+    }
+  }
+  
+  /* should we create subdimensions from source if not found in cache.
+  e.g. if dimension=mosaic returns dimension=val1,val2,val3 should we 
+  query the wms source with dimension=val1 , dimension=val2 and/or
+  dimension=val3 if they are not found in cache */
+  dimension_node = ezxml_child(node,"subdimensions_read_only");
+  if(dimension_node) {
+    if(tileset->dimension_assembly_type == MAPCACHE_DIMENSION_ASSEMBLY_NONE) {
+      ctx->set_error(ctx,400,"<subdimensions_read_only> used on a tileset with no <assembly_type> set, which makes no sense");
+      return;
+    }
+    if(dimension_node && dimension_node->txt && !strcmp(dimension_node->txt,"true")) {
+      tileset->subdimension_read_only = 1;
+    } else if(strcmp(dimension_node->txt,"false")) {
+      ctx->set_error(ctx,400,"failed to parse <subdimensions_read_only> (%s), expecting \"true\" or \"false\"",dimension_node->txt);
+      return;
+    }
+  }
 }
 
 void parseGrid(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
@@ -340,14 +336,14 @@ void parseSource(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
   source = NULL;
   if(!strcmp(type,"wms")) {
     source = mapcache_source_wms_create(ctx);
-#ifdef USE_MAPSERVER
   } else if(!strcmp(type,"mapserver")) {
     source = mapcache_source_mapserver_create(ctx);
-#endif
   } else if(!strcmp(type,"gdal")) {
     source = mapcache_source_gdal_create(ctx);
   } else if(!strcmp(type,"dummy")) {
     source = mapcache_source_dummy_create(ctx);
+  } else if(!strcmp(type,"fallback")) {
+    source = mapcache_source_fallback_create(ctx);
   } else {
     ctx->set_error(ctx, 400, "unknown source type %s for source \"%s\"", type, name);
     return;
@@ -362,8 +358,22 @@ void parseSource(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
     parseMetadata(ctx, cur_node, source->metadata);
     GC_CHECK_ERROR(ctx);
   }
+  if ((cur_node = ezxml_child(node,"retries")) != NULL) {
+    source->retry_count = atoi(cur_node->txt);
+    if(source->retry_count > 10) {
+      ctx->set_error(ctx,400,"source (%s) <retries> count of %d is unreasonably large. max is 10", source->name, source->retry_count);
+      return;
+    }
+  }
+  if ((cur_node = ezxml_child(node,"retry_delay")) != NULL) {
+    source->retry_delay = (double)atof(cur_node->txt);
+    if(source->retry_delay < 0) {
+      ctx->set_error(ctx,400,"source (%s) retry delay of %f must be positive",source->name, source->retry_delay);
+      return;
+    }
+  }
 
-  source->configuration_parse_xml(ctx,node,source);
+  source->configuration_parse_xml(ctx,node,source, config);
   GC_CHECK_ERROR(ctx);
   source->configuration_check(ctx,config,source);
   GC_CHECK_ERROR(ctx);
@@ -422,6 +432,7 @@ void parseFormat(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
     }
   } else if(!strcmp(type,"JPEG")) {
     int quality = 95;
+    int optimize = TRUE;
     mapcache_photometric photometric = MAPCACHE_PHOTOMETRIC_YCBCR;
     if ((cur_node = ezxml_child(node,"quality")) != NULL) {
       char *endptr;
@@ -445,10 +456,24 @@ void parseFormat(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
         return;
       }
     }
+    if ((cur_node = ezxml_child(node,"optimize")) != NULL) {
+      if(cur_node->txt && !strcasecmp(cur_node->txt,"false"))
+        optimize = MAPCACHE_OPTIMIZE_NO;
+      else if(cur_node->txt && !strcasecmp(cur_node->txt,"true"))
+        optimize = MAPCACHE_OPTIMIZE_YES;
+      else if(cur_node->txt && !strcasecmp(cur_node->txt,"arithmetic"))
+        optimize = MAPCACHE_OPTIMIZE_ARITHMETIC;
+      else {
+        ctx->set_error(ctx,500,"failed to parse jpeg format %s optimize %s. expecting true, false or arithmetic",
+                       name,cur_node->txt);
+        return;
+      }
+    }
     format = mapcache_imageio_create_jpeg_format(ctx->pool,
-             name,quality,photometric);
+             name,quality,photometric,optimize);
   } else if(!strcasecmp(type,"MIXED")) {
     mapcache_image_format *transparent=NULL, *opaque=NULL;
+    unsigned int alpha_cutoff=255;
     if ((cur_node = ezxml_child(node,"transparent")) != NULL) {
       transparent = mapcache_configuration_get_image_format(config,cur_node->txt);
     }
@@ -467,7 +492,10 @@ void parseFormat(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
                      name,cur_node->txt,cur_node->txt);
       return;
     }
-    format = mapcache_imageio_create_mixed_format(ctx->pool,name,transparent, opaque);
+    if ((cur_node = ezxml_child(node,"alpha_cutoff")) != NULL) {
+      alpha_cutoff = atoi(cur_node->txt);
+    }
+    format = mapcache_imageio_create_mixed_format(ctx->pool,name,transparent, opaque, alpha_cutoff);
   } else {
     ctx->set_error(ctx, 400, "unknown format type %s for format \"%s\"", type, name);
     return;
@@ -486,6 +514,7 @@ void parseCache(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
 {
   char *name = NULL,  *type = NULL;
   mapcache_cache *cache = NULL;
+  ezxml_t cur_node;
   name = (char*)ezxml_attr(node,"name");
   type = (char*)ezxml_attr(node,"type");
   if(!name || !strlen(name)) {
@@ -520,69 +549,48 @@ void parseCache(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
   } else if(!strcmp(type,"google")) {
     cache = mapcache_cache_google_create(ctx);
   } else if(!strcmp(type,"bdb")) {
-#ifdef USE_BDB
     cache = mapcache_cache_bdb_create(ctx);
-#else
-    ctx->set_error(ctx,400, "failed to add cache \"%s\": Berkeley DB support is not available on this build",name);
-    return;
-#endif
   } else if(!strcmp(type,"tokyocabinet")) {
-#ifdef USE_TC
     cache = mapcache_cache_tc_create(ctx);
-#else
-    ctx->set_error(ctx,400, "failed to add cache \"%s\": Tokyo Cabinet support is not available on this build",name);
-    return;
-#endif
   } else if(!strcmp(type,"sqlite3")) {
-#ifdef USE_SQLITE
     cache = mapcache_cache_sqlite_create(ctx);
-#else
-    ctx->set_error(ctx,400, "failed to add cache \"%s\": sqlite support is not available on this build",name);
-    return;
-#endif
   } else if(!strcmp(type,"mbtiles")) {
-#ifdef USE_SQLITE
     cache = mapcache_cache_mbtiles_create(ctx);
-#else
-    ctx->set_error(ctx,400, "failed to add cache \"%s\": sqlite support is not available on this build",name);
-    return;
-#endif
   } else if(!strcmp(type,"memcache")) {
-#ifdef USE_MEMCACHE
     cache = mapcache_cache_memcache_create(ctx);
-#else
-    ctx->set_error(ctx,400, "failed to add cache \"%s\": memcache support is not available on this build",name);
-    return;
-#endif
   } else if(!strcmp(type,"tiff")) {
-#ifdef USE_TIFF
     cache = mapcache_cache_tiff_create(ctx);
-#else
-    ctx->set_error(ctx,400, "failed to add cache \"%s\": tiff support is not available on this build",name);
-    return;
-#endif
   } else if(!strcmp(type,"couchbase")) {
-#ifdef USE_COUCHBASE
     cache = mapcache_cache_couchbase_create(ctx);
-#else
-    ctx->set_error(ctx, 400, "failed to add cache \"%s\": couchbase support is not available on this build", name);
-#endif
   } else if(!strcmp(type,"riak")) {
-#ifdef USE_RIAK
     cache = mapcache_cache_riak_create(ctx);
-#else
-    ctx->set_error(ctx, 400, "failed to add cache \"%s\": riak support is not available on this build", name);
-#endif
   } else {
     ctx->set_error(ctx, 400, "unknown cache type %s for cache \"%s\"", type, name);
     return;
   }
+  GC_CHECK_ERROR(ctx);
   if(cache == NULL) {
     ctx->set_error(ctx, 400, "failed to parse cache \"%s\"", name);
     return;
   }
   cache->name = name;
 
+  if ((cur_node = ezxml_child(node,"retries")) != NULL) {
+    cache->retry_count = atoi(cur_node->txt);
+    if(cache->retry_count > 10) {
+      ctx->set_error(ctx,400,"cache (%s) <retries> count of %d is unreasonably large. max is 10", cache->name, cache->retry_count);
+      return;
+    }
+  }
+  if ((cur_node = ezxml_child(node,"retry_delay")) != NULL) {
+    cache->retry_delay = (double)atof(cur_node->txt);
+    if(cache->retry_delay < 0) {
+      ctx->set_error(ctx,400,"cache (%s) retry delay of %f must be positive",cache->name, cache->retry_delay);
+      return;
+    }
+  }
+
+
   cache->configuration_parse_xml(ctx,node,cache,config);
   GC_CHECK_ERROR(ctx);
   mapcache_configuration_add_cache(config,cache,name);
@@ -824,12 +832,6 @@ void parseTileset(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
     GC_CHECK_ERROR(ctx);
   }
   
-  if ((cur_node = ezxml_child(node,"timedimension")) != NULL) {
-    parseTimeDimension(ctx, cur_node, tileset);
-    GC_CHECK_ERROR(ctx);
-  }
-
-
   if ((cur_node = ezxml_child(node,"cache")) != NULL) {
     mapcache_cache *cache = mapcache_configuration_get_cache(config, cur_node->txt);
     if(!cache) {
@@ -1115,8 +1117,6 @@ void mapcache_configuration_parse_xml(mapcache_context *ctx, const char *filenam
   } else if ((node = ezxml_child(doc,"services")) != NULL) {
     ctx->log(ctx,MAPCACHE_WARN,"<services> tag is deprecated, use <service type=\"wms\" enabled=\"true|false\">");
     parseServices(ctx, node, config);
-  } else {
-    ctx->set_error(ctx, 400, "no <services> configured");
   }
   if(GC_HAS_ERROR(ctx)) goto cleanup;
 
@@ -1158,31 +1158,8 @@ void mapcache_configuration_parse_xml(mapcache_context *ctx, const char *filenam
     mapcache_config_parse_locker(ctx,node,&config->locker);
     GC_CHECK_ERROR(ctx);
   } else {
-    /* backwards compatibility */
-    int micro_retry;
-    mapcache_locker_disk *ldisk;
-    config->locker = mapcache_locker_disk_create(ctx);
-    ldisk = (mapcache_locker_disk*)config->locker;
-    if((node = ezxml_child(doc,"lock_dir")) != NULL) {
-      ldisk->dir = apr_pstrdup(ctx->pool, node->txt);
-    } else {
-      ldisk->dir = apr_pstrdup(ctx->pool,"/tmp");
-    }
-
-    if((node = ezxml_child(doc,"lock_retry")) != NULL) {
-      char *endptr;
-      micro_retry = strtol(node->txt,&endptr,10);
-      if(*endptr != 0 || micro_retry <= 0) {
-        ctx->set_error(ctx, 400, "failed to parse lock_retry microseconds \"%s\". Expecting a positive integer",
-            node->txt);
-        return;
-      }
-    } else {
-      /* default retry interval is 1/100th of a second, i.e. 10000 microseconds */
-      micro_retry = 10000;
-    }
-    config->locker->retry_interval = micro_retry / 1000000.0;
-    config->locker->timeout=120;
+    mapcache_config_parse_locker_old(ctx,doc,config);
+    GC_CHECK_ERROR(ctx);
   }
 
   if((node = ezxml_child(doc,"threaded_fetching")) != NULL) {
