@@ -71,7 +71,7 @@ mapcache_tileset *tileset;
 mapcache_tileset *tileset_transfer;
 mapcache_cfg *cfg;
 mapcache_context ctx;
-apr_table_t *dimensions=NULL;
+apr_array_header_t *dimensions=NULL;
 int minzoom=-1;
 int maxzoom=-1;
 mapcache_grid_link *grid_link;
@@ -83,7 +83,9 @@ int force = 0;
 int sig_int_received = 0;
 int error_detected = 0;
 double percent_failed_allowed = 1.0;
-int n_metatiles_tot=0;
+int n_metatiles_tot = 0;
+int n_nodata_tot = 0;
+int rate_limit = 0;
 FILE *failed_log = NULL, *retry_log = NULL;
 #define FAIL_BACKLOG_COUNT 1000
 
@@ -95,7 +97,8 @@ typedef enum {
   MAPCACHE_CMD_STOP,
   MAPCACHE_CMD_DELETE,
   MAPCACHE_CMD_SKIP,
-  MAPCACHE_CMD_TRANSFER
+  MAPCACHE_CMD_TRANSFER,
+  MAPCACHE_CMD_STOP_RECURSION
 } cmd;
 
 typedef enum {
@@ -123,6 +126,7 @@ typedef enum {
 struct seed_status {
   s_status status;
   int x,y,z;
+  int nodata;
   char *msg;
 };
 
@@ -138,6 +142,8 @@ cmd mode = MAPCACHE_CMD_SEED; /* the mode the utility will be running in: either
 int push_queue(struct seed_cmd cmd)
 {
   struct seed_cmd *pcmd;
+  int retries=0;
+  int ret;
 #ifdef USE_FORK
   if(nprocesses > 1) {
     struct msg_cmd mcmd;
@@ -152,12 +158,21 @@ int push_queue(struct seed_cmd cmd)
 #endif
   pcmd = calloc(1,sizeof(struct seed_cmd));
   *pcmd = cmd;
-  return apr_queue_push(work_queue,pcmd);
+  ret = apr_queue_push(work_queue,pcmd);
+  while (ret == APR_EINTR && retries < 10) {
+    retries++;
+    ret = apr_queue_push(work_queue,pcmd);
+  }
+  if(ret == APR_EINTR) {
+    printf("failed to push tile %d %d %d after 10 retries\n",cmd.z,cmd.y,cmd.x);
+    return APR_EGENERAL;
+  }
+  return ret;
 }
 
 int pop_queue(struct seed_cmd *cmd)
 {
-  int ret;
+  int ret,retries=0;
   struct seed_cmd *pcmd;
 
 #ifdef USE_FORK
@@ -173,6 +188,10 @@ int pop_queue(struct seed_cmd *cmd)
 #endif
 
   ret = apr_queue_pop(work_queue, (void**)&pcmd);
+  while(ret == APR_EINTR && retries<10) {
+    retries++;
+    ret = apr_queue_pop(work_queue, (void**)&pcmd);
+  }
   if(ret == APR_SUCCESS) {
     *cmd = *pcmd;
     free(pcmd);
@@ -182,7 +201,7 @@ int pop_queue(struct seed_cmd *cmd)
 
 int trypop_queue(struct seed_cmd *cmd)
 {
-  int ret;
+  int ret,retries=0;
   struct seed_cmd *pcmd;
 
 #ifdef USE_FORK
@@ -200,6 +219,10 @@ int trypop_queue(struct seed_cmd *cmd)
   }
 #endif
   ret = apr_queue_trypop(work_queue,(void**)&pcmd);
+  while(ret == APR_EINTR && retries<10) {
+    retries++;
+    ret = apr_queue_trypop(work_queue,(void**)&pcmd);
+  }
   if(ret == APR_SUCCESS) {
     *cmd = *pcmd;
     free(pcmd);
@@ -207,6 +230,9 @@ int trypop_queue(struct seed_cmd *cmd)
   return ret;
 }
 
+#define SEEDER_OPT_THREAD_DELAY 256
+#define SEEDER_OPT_RATE_LIMIT 257
+
 static const apr_getopt_option_t seed_options[] = {
   /* long-option, short-option, has-arg flag, description */
   { "config", 'c', TRUE, "configuration file (/path/to/mapcache.xml)"},
@@ -219,7 +245,7 @@ static const apr_getopt_option_t seed_options[] = {
   { "force", 'f', FALSE, "force tile recreation even if it already exists" },
   { "grid", 'g', TRUE, "grid to seed" },
   { "help", 'h', FALSE, "show help" },
-  { "iteration-mode", 'i', TRUE, "either \"drill-down\" or \"level-by-level\". Default is to use drill-down for g, WGS84 and GoogleMapsCompatible grids, and level-by-level for others. Use this flag to override." },
+  { "iteration-mode", 'i', TRUE, "either \"drill-down\" or \"scanline\". Default is to use drill-down for g, WGS84 and GoogleMapsCompatible grids, and scanline for others. Use this flag to override." },
 #ifdef USE_CLIPPERS
   { "ogr-layer", 'l', TRUE, "layer inside datasource"},
 #endif
@@ -242,14 +268,19 @@ static const apr_getopt_option_t seed_options[] = {
 #endif
   { "transfer", 'x', TRUE, "tileset to transfer" },
   { "zoom", 'z', TRUE, "min and max zoomlevels to seed, separated by a comma. eg 0,6" },
+  { "rate-limit", SEEDER_OPT_RATE_LIMIT, TRUE, "maximum number of tiles/second to seed"},
+  { "thread-delay", SEEDER_OPT_THREAD_DELAY, TRUE, "delay in seconds between rendering thread creation (ramp up)"},
   { NULL, 0, 0, NULL }
 };
 
 void handle_sig_int(int signal)
 {
+#define signal_msg "SIGINT received, waiting for threads to finish\npress ctrl-C again to force terminate\n"
   if(!sig_int_received) {
-    fprintf(stderr,"SIGINT received, waiting for threads to finish\n");
-    fprintf(stderr,"press ctrl-C again to force terminate, you might end up with locked tiles\n");
+    int err = write(2,signal_msg,strlen(signal_msg));
+    if(err) {
+      //nothing we can really do here
+    }
     sig_int_received = 1;
   } else {
     exit(signal);
@@ -258,12 +289,12 @@ void handle_sig_int(int signal)
 
 void seed_log(mapcache_context *ctx, mapcache_log_level level, char *msg, ...)
 {
-  if(verbose) {
+  if(level >= MAPCACHE_WARN || verbose) {
     va_list args;
     va_start(args,msg);
     vfprintf(stderr,msg,args);
     va_end(args);
-    printf("\n");
+    fprintf(stderr,"\n");
   }
 }
 
@@ -274,7 +305,7 @@ void mapcache_context_seeding_log(mapcache_context *ctx, mapcache_log_level leve
   va_start(args,msg);
   vfprintf(stderr,msg,args);
   va_end(args);
-  printf("\n");
+  fprintf(stderr,"\n");
 }
 
 #ifdef USE_CLIPPERS
@@ -317,31 +348,68 @@ cmd examine_tile(mapcache_context *ctx, mapcache_tile *tile)
 #ifdef USE_CLIPPERS
   /* check we are in the requested features before checking the tile */
   if(nClippers > 0 && ogr_features_intersect_tile(ctx,tile) == 0)
-    return MAPCACHE_CMD_SKIP;
+    return MAPCACHE_CMD_STOP_RECURSION;
 #endif
 
-  tile_exists = force?0:tileset->_cache->tile_exists(ctx,tileset->_cache,tile);
+  if(mode != MAPCACHE_CMD_TRANSFER && force) {
+    if(mode == MAPCACHE_CMD_DELETE) {
+      tile_exists = 1;
+    } else {
+      tile_exists = 0;
+    }
+  } else {
+    int i;
+    if(tile->tileset->dimension_assembly_type != MAPCACHE_DIMENSION_ASSEMBLY_NONE) {
+      for(i=0; i<tile->dimensions->nelts; i++) {
+        mapcache_requested_dimension *rdim = APR_ARRAY_IDX(tile->dimensions,i,mapcache_requested_dimension*);
+        rdim->cached_value = rdim->requested_value;
+      }
+    } else {
+      if(tile->dimensions) {
+        mapcache_extent extent;
+        mapcache_grid_get_tile_extent(ctx,tile->grid_link->grid,tile->x,tile->y,tile->z,&extent);
+        for(i=0; i<tile->dimensions->nelts; i++) {
+          apr_array_header_t *rdim_vals;
+          mapcache_requested_dimension *rdim = APR_ARRAY_IDX(tile->dimensions,i,mapcache_requested_dimension*);
+          rdim_vals = rdim->dimension->get_entries_for_value(ctx,rdim->dimension,rdim->requested_value, tile->tileset, NULL, tile->grid_link->grid);
+          if(GC_HAS_ERROR(ctx)) {
+            return MAPCACHE_CMD_SKIP;
+          }
+          if(rdim_vals->nelts > 1) {
+            ctx->set_error(ctx,500,"dimension (%s) for tileset (%s) returned invalid number of subdimensions (1 expected)",rdim->dimension->name, tile->tileset->name);
+            return MAPCACHE_CMD_SKIP;
+          }
+          if(rdim_vals->nelts == 0) {
+            ctx->set_error(ctx,404,"dimension (%s) for tileset (%s) returned no subdimensions (1 expected)",rdim->dimension->name, tile->tileset->name);
+            return MAPCACHE_CMD_SKIP;
+          }
+          rdim->cached_value = APR_ARRAY_IDX(rdim_vals,0,char*);
+        }
+      }
+    }
+    tile_exists = mapcache_cache_tile_exists(ctx,tileset->_cache,tile);
+  }
 
   /* if the tile exists and a time limit was specified, check the tile modification date */
   if(tile_exists) {
     if(age_limit) {
-      if(tileset->_cache->tile_get(ctx,tileset->_cache, tile) == MAPCACHE_SUCCESS) {
+      if(mapcache_cache_tile_get(ctx,tileset->_cache, tile) == MAPCACHE_SUCCESS) {
         if(tile->mtime && tile->mtime<age_limit) {
           /* the tile modification time is older than the specified limit */
-      if(mode == MAPCACHE_CMD_SEED || mode == MAPCACHE_CMD_TRANSFER) {
-        mapcache_tileset_tile_delete(ctx,tile,MAPCACHE_TRUE);
-        /* if we are in mode transfer, delete it from the dst tileset */
-        if (mode == MAPCACHE_CMD_TRANSFER) {
-          tile->tileset = tileset_transfer;
-          if (tile->tileset->_cache->tile_exists(ctx,tile->tileset->_cache, tile)) {
-        mapcache_tileset_tile_delete(ctx,tile,MAPCACHE_TRUE);
+          if(mode == MAPCACHE_CMD_SEED || mode == MAPCACHE_CMD_TRANSFER) {
+            mapcache_tileset_tile_delete(ctx,tile,MAPCACHE_TRUE);
+            /* if we are in mode transfer, delete it from the dst tileset */
+            if (mode == MAPCACHE_CMD_TRANSFER) {
+              tile->tileset = tileset_transfer;
+              if (mapcache_cache_tile_exists(ctx,tile->tileset->_cache, tile)) {
+                mapcache_tileset_tile_delete(ctx,tile,MAPCACHE_TRUE);
+              }
+              tile->tileset = tileset;
+            }
+            action = mode;
+          } else { //if(action == MAPCACHE_CMD_DELETE)
+            action = MAPCACHE_CMD_DELETE;
           }
-          tile->tileset = tileset;
-        }
-        action = mode;
-      } else { //if(action == MAPCACHE_CMD_DELETE)
-        action = MAPCACHE_CMD_DELETE;
-      }
         }
       } else {
         //BUG: tile_exists returned true, but tile_get returned a failure. not sure what to do.
@@ -355,7 +423,7 @@ cmd examine_tile(mapcache_context *ctx, mapcache_tile *tile)
         /* the tile exists in the source tileset,
            check if the tile exists in the destination cache */
         tile->tileset = tileset_transfer;
-        if (tile->tileset->_cache->tile_exists(ctx,tile->tileset->_cache, tile)) {
+        if (!force && mapcache_cache_tile_exists(ctx,tile->tileset->_cache, tile)) {
           action = MAPCACHE_CMD_SKIP;
         } else {
           action = MAPCACHE_CMD_TRANSFER;
@@ -368,7 +436,7 @@ cmd examine_tile(mapcache_context *ctx, mapcache_tile *tile)
     }
   } else {
     // the tile does not exist
-    if(mode == MAPCACHE_CMD_SEED || mode == MAPCACHE_CMD_TRANSFER) {
+    if (mode == MAPCACHE_CMD_SEED) {
       action = mode;
     } else {
       action = MAPCACHE_CMD_SKIP;
@@ -378,6 +446,27 @@ cmd examine_tile(mapcache_context *ctx, mapcache_tile *tile)
   return action;
 }
 
+double rate_limit_last_time = 0;
+double rate_limit_delay = 0.0;
+
+void rate_limit_sleep() {
+  if(rate_limit > 0) {
+    struct mctimeval now;
+    double now_time;
+    mapcache_gettimeofday(&now,NULL);
+    now_time = now.tv_sec + now.tv_usec / 1000000.0;
+
+    if((now_time - rate_limit_last_time) < rate_limit_delay) {
+      apr_sleep((int)((rate_limit_delay - (now_time - rate_limit_last_time)) * 1000000));
+      /*last_time = now_time + delay - (now_time - last_time); //now plus the time we slept */
+      rate_limit_last_time = rate_limit_delay + rate_limit_last_time;
+    } else {
+      rate_limit_last_time = now_time;
+    }
+
+  }
+}
+
 void cmd_recurse(mapcache_context *cmd_ctx, mapcache_tile *tile)
 {
   cmd action;
@@ -404,8 +493,13 @@ void cmd_recurse(mapcache_context *cmd_ctx, mapcache_tile *tile)
     cmd.y = tile->y;
     cmd.z = tile->z;
     cmd.command = action;
+    if(rate_limit > 0)
+      rate_limit_sleep();
     push_queue(cmd);
   }
+  
+  if(action == MAPCACHE_CMD_STOP_RECURSION)
+    return;
 
   //recurse into our 4 child metatiles
 
@@ -424,9 +518,9 @@ void cmd_recurse(mapcache_context *cmd_ctx, mapcache_tile *tile)
    */
 
 
-  mapcache_grid_get_extent(cmd_ctx, grid_link->grid,
+  mapcache_grid_get_tile_extent(cmd_ctx, grid_link->grid,
                            curx, cury, curz, &bboxbl);
-  mapcache_grid_get_extent(cmd_ctx, grid_link->grid,
+  mapcache_grid_get_tile_extent(cmd_ctx, grid_link->grid,
                            curx+tileset->metasize_x-1, cury+tileset->metasize_y-1, curz, &bboxtr);
   epsilon = (bboxbl.maxx-bboxbl.minx)*0.01;
   mapcache_grid_get_xy(cmd_ctx,grid_link->grid,
@@ -458,7 +552,7 @@ void cmd_recurse(mapcache_context *cmd_ctx, mapcache_tile *tile)
   tile->z = curz;
 }
 
-void cmd_worker()
+void feed_worker()
 {
   int n;
   mapcache_tile *tile;
@@ -470,7 +564,11 @@ void cmd_worker()
   if(nprocesses >= 1) nworkers = nprocesses;
   apr_pool_create(&cmd_ctx.pool,ctx.pool);
   tile = mapcache_tileset_tile_create(ctx.pool, tileset, grid_link);
-  tile->dimensions = dimensions;
+  tile->dimensions = mapcache_requested_dimensions_clone(ctx.pool,dimensions);
+  if(rate_limit > 0) {
+    /* compute time between seed commands accounting for max rate-limit and current metasize */
+    rate_limit_delay = (tileset->metasize_x * tileset->metasize_y) / (double)rate_limit;
+  }
   if(iteration_mode == MAPCACHE_ITERATION_DEPTH_FIRST) {
     do {
       tile->x = x;
@@ -518,6 +616,8 @@ void cmd_worker()
         cmd.y = y;
         cmd.z = z;
         cmd.command = action;
+        if(rate_limit > 0)
+          rate_limit_sleep();
         push_queue(cmd);
       }
 
@@ -555,7 +655,9 @@ void seed_worker()
   apr_pool_create(&seed_ctx.pool,ctx.pool);
   apr_pool_create(&tpool,ctx.pool);
   tile = mapcache_tileset_tile_create(tpool, tileset, grid_link);
-  tile->dimensions = dimensions;
+  if(dimensions) {
+    tile->dimensions = mapcache_requested_dimensions_clone(tpool,dimensions);
+  }
   while(1) {
     struct seed_cmd cmd;
     apr_status_t ret;
@@ -566,35 +668,51 @@ void seed_worker()
     tile->x = cmd.x;
     tile->y = cmd.y;
     tile->z = cmd.z;
-    if(cmd.command == MAPCACHE_CMD_SEED) {
-      /* aquire a lock on the metatile ?*/
-      mapcache_metatile *mt = mapcache_tileset_metatile_get(&seed_ctx, tile);
-      void *lock;
-      int isLocked = mapcache_lock_or_wait_for_resource(&seed_ctx, seed_ctx.config->locker, mapcache_tileset_metatile_resource_key(&seed_ctx,mt), &lock);
-      if(isLocked == MAPCACHE_TRUE) {
-        /* this will query the source to create the tiles, and save them to the cache */
-        mapcache_tileset_render_metatile(&seed_ctx, mt);
-        if(GC_HAS_ERROR(&seed_ctx)) {
-          /* temporarily clear error state so we don't mess up with error handling in the locker */
-          void *error;
-          seed_ctx.pop_errors(&seed_ctx,&error);
-          mapcache_unlock_resource(&seed_ctx, seed_ctx.config->locker, mapcache_tileset_metatile_resource_key(&seed_ctx,mt), lock);
-          seed_ctx.push_errors(&seed_ctx,error);
-        } else {
-          mapcache_unlock_resource(&seed_ctx, seed_ctx.config->locker, mapcache_tileset_metatile_resource_key(&seed_ctx,mt), lock);
+    tile->nodata = 0;
+    tile->encoded_data = NULL;
+    tile->raw_image = NULL;
+    if(tile->dimensions) {
+      int i;
+      if(tileset->dimension_assembly_type == MAPCACHE_DIMENSION_ASSEMBLY_NONE) {
+        mapcache_extent extent;
+        mapcache_grid_get_tile_extent(&seed_ctx,tile->grid_link->grid,tile->x,tile->y,tile->z,&extent);
+        for(i=0; i<tile->dimensions->nelts; i++) {
+          apr_array_header_t *rdim_vals;
+          mapcache_requested_dimension *rdim = APR_ARRAY_IDX(tile->dimensions,i,mapcache_requested_dimension*);
+          rdim_vals = rdim->dimension->get_entries_for_value(&seed_ctx,rdim->dimension,rdim->requested_value, tile->tileset, NULL, tile->grid_link->grid);
+          GC_CHECK_ERROR(&seed_ctx);
+          if(rdim_vals->nelts > 1) {
+            seed_ctx.set_error(&seed_ctx,500,"dimension (%s) for tileset (%s) returned invalid number of subdimensions (1 expected)",rdim->dimension->name, tile->tileset->name);
+            return;
+          }
+          if(rdim_vals->nelts == 0) {
+            seed_ctx.set_error(&seed_ctx,404,"dimension (%s) for tileset (%s) returned no subdimensions (1 expected)",rdim->dimension->name, tile->tileset->name);
+            return;
+          }
+          rdim->cached_value = APR_ARRAY_IDX(rdim_vals,0,char*);
+        }
+      } else {
+        for(i=0; i<tile->dimensions->nelts; i++) {
+          mapcache_requested_dimension *rdim = APR_ARRAY_IDX(tile->dimensions,i,mapcache_requested_dimension*);
+          rdim->cached_value = NULL;
         }
       }
+    }
+    if(cmd.command == MAPCACHE_CMD_SEED) {
+      if(!tile->dimensions || tileset->dimension_assembly_type == MAPCACHE_DIMENSION_ASSEMBLY_NONE) {
+        mapcache_metatile *mt = mapcache_tileset_metatile_get(&seed_ctx, tile);
+        /* this will query the source to create the tiles, and save them to the cache */
+        mapcache_tileset_render_metatile(&seed_ctx, mt);
+      } else {
+        mapcache_tileset_tile_set_get_with_subdimensions(&seed_ctx,tile);
+      }
     } else if (cmd.command == MAPCACHE_CMD_TRANSFER) {
-      int i;
-      mapcache_metatile *mt = mapcache_tileset_metatile_get(&seed_ctx, tile);
-      for (i = 0; i < mt->ntiles; i++) {
-        mapcache_tile *subtile = &mt->tiles[i];
-        int cache_ret;
-        cache_ret = tileset->_cache->tile_get(&seed_ctx, tileset->_cache, subtile);
-        if(cache_ret == MAPCACHE_SUCCESS && !GC_HAS_ERROR(&seed_ctx)) {
-          subtile->tileset = tileset_transfer;
-          subtile->tileset->_cache->tile_set(&seed_ctx, subtile->tileset->_cache, subtile);
-        }
+      mapcache_tileset_tile_get(&seed_ctx, tile);
+      if(!tile->nodata && !GC_HAS_ERROR(&seed_ctx)) {
+        mapcache_tileset *tmp_tileset = tile->tileset;
+        tile->tileset = tileset_transfer;
+        mapcache_cache_tile_set(&seed_ctx, tile->tileset->_cache, tile);
+        tile->tileset = tmp_tileset;
       }
     } else { //CMD_DELETE
       mapcache_tileset_tile_delete(&seed_ctx,tile,MAPCACHE_TRUE);
@@ -602,9 +720,11 @@ void seed_worker()
 
     {
       struct seed_status *st = calloc(1,sizeof(struct seed_status));
+      int retries=0;
       st->x=tile->x;
       st->y=tile->y;
       st->z=tile->z;
+      st->nodata = tile->nodata;
       if(seed_ctx.get_error(&seed_ctx)) {
         st->status = MAPCACHE_STATUS_FAIL;
         st->msg = strdup(seed_ctx.get_error_message(&seed_ctx));
@@ -613,6 +733,14 @@ void seed_worker()
         st->status = MAPCACHE_STATUS_OK;
       }
       ret = apr_queue_push(log_queue,(void*)st);
+      while( ret == APR_EINTR && retries < 10) {
+        retries++;
+        ret = apr_queue_push(log_queue,(void*)st);
+      }
+      if( ret == APR_EINTR) {
+        printf("FATAL ERROR: unable to log progress after 10 retries, aborting\n");
+        break;
+      }
       if(ret != APR_SUCCESS)
       {
         printf("FATAL ERROR: unable to log progress\n");
@@ -628,11 +756,18 @@ int seed_process() {
   return 0;
 }
 #endif
+
 static void* APR_THREAD_FUNC seed_thread(apr_thread_t *thread, void *data) {
   seed_worker();
   return NULL;
 }
 
+static void* APR_THREAD_FUNC feed_thread_fn(apr_thread_t *thread, void *data) {
+  //start the thread that will populate the queue.
+  feed_worker();
+  return NULL;
+}
+
 static void* APR_THREAD_FUNC log_thread_fn(apr_thread_t *thread, void *data) {
   size_t cur;
   double last_time;
@@ -646,14 +781,22 @@ static void* APR_THREAD_FUNC log_thread_fn(apr_thread_t *thread, void *data) {
   cur=0;
   last_time=0;
   while(1) {
+    int retries = 0;
     struct seed_status *st;
     apr_status_t ret = apr_queue_pop(log_queue, (void**)&st);
+    while(ret == APR_EINTR && retries<10) {
+      retries++;
+      ret = apr_queue_pop(log_queue, (void**)&st);
+    }
     if(ret != APR_SUCCESS || !st) break;
     if(st->status == MAPCACHE_STATUS_FINISHED)
       return NULL;
     if(st->status == MAPCACHE_STATUS_OK) {
       failed[cur]=0;
       n_metatiles_tot++;
+      if(st->nodata) {
+        n_nodata_tot++;
+      }
       if(!quiet) {
         struct mctimeval now;
         mapcache_gettimeofday(&now,NULL);
@@ -682,7 +825,6 @@ static void* APR_THREAD_FUNC log_thread_fn(apr_thread_t *thread, void *data) {
       if(pct > percent_failed_allowed) {
         ctx.log(&ctx, MAPCACHE_ERROR, "aborting seed as %.1f%% of the last %d requests failed\n", pct, FAIL_BACKLOG_COUNT);
         error_detected = 1;
-        return NULL;
       }
     }
     if(st->msg) free(st->msg);
@@ -736,10 +878,18 @@ int usage(const char *progname, char *msg, ...)
     printf("usage: %s options\n",progname);
 
   while(seed_options[i].name) {
-    if(seed_options[i].has_arg==TRUE) {
-      printf("-%c|--%s [value]: %s\n",seed_options[i].optch,seed_options[i].name, seed_options[i].description);
+    if(seed_options[i].optch<256) {
+      if(seed_options[i].has_arg==TRUE) {
+        printf("-%c|--%s [value]: %s\n",seed_options[i].optch,seed_options[i].name, seed_options[i].description);
+      } else {
+        printf("-%c|--%s: %s\n",seed_options[i].optch,seed_options[i].name, seed_options[i].description);
+      }
     } else {
-      printf("-%c|--%s: %s\n",seed_options[i].optch,seed_options[i].name, seed_options[i].description);
+      if(seed_options[i].has_arg==TRUE) {
+        printf("   --%s [value]: %s\n",seed_options[i].name, seed_options[i].description);
+      } else {
+        printf("   --%s: %s\n",seed_options[i].name, seed_options[i].description);
+      }
     }
     i++;
   }
@@ -757,9 +907,8 @@ int main(int argc, const char **argv)
   /* initialize apr_getopt_t */
   apr_getopt_t *opt;
   const char *configfile=NULL;
-  apr_thread_t **threads;
-  apr_thread_t *log_thread;
-  apr_threadattr_t *thread_attrs;
+  apr_thread_t **seed_threads;
+  apr_thread_t *log_thread,*feed_thread;
   const char *tileset_name=NULL;
   const char *tileset_transfer_name=NULL;
   const char *grid_name = NULL;
@@ -776,6 +925,7 @@ int main(int argc, const char **argv)
   int *metasizes = NULL;//[2];
   int metax=-1,metay=-1;
   double *extent_array = NULL;
+  double thread_delay = 0.0;
 
 #ifdef USE_CLIPPERS
   OGRFeatureH hFeature;
@@ -832,10 +982,10 @@ int main(int argc, const char **argv)
       case 'i':
         if(!strcmp(optarg,"drill-down")) {
           iteration_mode = MAPCACHE_ITERATION_DEPTH_FIRST;
-        } else if(!strcmp(optarg,"level-by-level")) {
+        } else if(!strcmp(optarg,"level-by-level") || !strcmp(optarg, "scanline")) {
           iteration_mode = MAPCACHE_ITERATION_LEVEL_FIRST;
         } else {
-          return usage(argv[0],"invalid iteration mode, expecting \"drill-down\" or \"level-by-level\"");
+          return usage(argv[0],"invalid iteration mode, expecting \"drill-down\" or \"scanline\"");
         }
         break;
       case 'L':
@@ -930,6 +1080,16 @@ int main(int argc, const char **argv)
         }
         apr_table_set(argdimensions,dimkey,dimvalue);
         break;
+      case SEEDER_OPT_THREAD_DELAY:
+        thread_delay = strtod(optarg, NULL);
+        if(thread_delay < 0.0 )
+          return usage(argv[0], "failed to parse thread-delay, expecting positive number of seconds");
+        break;
+      case SEEDER_OPT_RATE_LIMIT:
+        rate_limit = (int)strtol(optarg, NULL, 10);
+        if(rate_limit <= 0 )
+          return usage(argv[0], "failed to parse rate-limit, expecting positive number of tiles per seconds");
+        break;
 #ifdef USE_CLIPPERS
       case 'd':
         ogr_datasource = optarg;
@@ -1064,10 +1224,16 @@ int main(int argc, const char **argv)
     if(!tileset) {
       return usage(argv[0], "tileset not found in configuration");
     }
-    if(tileset->read_only) {
-      printf("tileset %s is read-only, switching it to read-write for seeding\n",tileset_name);
+    if(tileset->read_only && mode != MAPCACHE_CMD_TRANSFER) {
+      printf("tileset (%s) is read-only, switching it to read-write for seeding\n",tileset_name);
       tileset->read_only = 0;
     }
+    if( mode == MAPCACHE_CMD_TRANSFER ) {
+      if( tileset->read_only == 0 ) {
+        printf("switching tileset (%s) to read-only as we are in transfer mode\n",tileset_name);
+      }
+      tileset->read_only = 1;
+    }
     if( ! grid_name ) {
       grid_link = APR_ARRAY_IDX(tileset->grid_links,0,mapcache_grid_link*);
     } else {
@@ -1134,7 +1300,7 @@ int main(int argc, const char **argv)
     if(minzoom<grid_link->minz) minzoom = grid_link->minz;
     if(maxzoom>= grid_link->maxz) maxzoom = grid_link->maxz - 1;
     if(grid_link->outofzoom_strategy != MAPCACHE_OUTOFZOOM_NOTCONFIGURED && maxzoom > grid_link->max_cached_zoom) {
-      maxzoom = grid_link->max_cached_zoom;
+      return usage(argv[0],"requested maxzoom %d is higher than configured max-cached-zoom %d for grid",maxzoom,grid_link->max_cached_zoom);
     }
 
     /* adjust metasize */
@@ -1146,14 +1312,14 @@ int main(int argc, const char **argv)
     /* ensure our metasize is a power of 2 in drill down mode */
     if(iteration_mode == MAPCACHE_ITERATION_DEPTH_FIRST) {
       if(!isPowerOfTwo(tileset->metasize_x) || !isPowerOfTwo(tileset->metasize_y)) {
-        return usage(argv[0],"metatile size is not set to a power of two and iteration mode set to \"drill-down\", rerun with e.g -M 8,8, or force iteration mode to \"level-by-level\"");
+        return usage(argv[0],"metatile size is not set to a power of two and iteration mode set to \"drill-down\", rerun with e.g -M 8,8, or force iteration mode to \"scanline\"");
       }
     }
 
     if(cache_override) {
       mapcache_cache *co = mapcache_configuration_get_cache(cfg, cache_override);
       if(!co) {
-        return usage(argv[0], "overrided cache\"%s\" to not found in configuration", cache_override);
+        return usage(argv[0], "overridden cache\"%s\" not found in configuration", cache_override);
       } else {
         tileset->_cache = co;
       }
@@ -1162,6 +1328,7 @@ int main(int argc, const char **argv)
   }
 
   if (mode == MAPCACHE_CMD_TRANSFER) {
+    tileset->metasize_x = tileset->metasize_y = 1;
     if (!tileset_transfer_name)
       return usage(argv[0],"tileset where tiles should be transferred to not specified");
 
@@ -1211,61 +1378,30 @@ int main(int argc, const char **argv)
   }
 
   /* validate the supplied dimensions */
-  if (!apr_is_empty_array(tileset->dimensions) || tileset->timedimension) {
+  if (!apr_is_empty_array(tileset->dimensions)) {
     int i;
     const char *value;
-    dimensions = apr_table_make(ctx.pool,3);
-    if (!apr_is_empty_array(tileset->dimensions)) {
-      for(i=0; i<tileset->dimensions->nelts; i++) {
-        mapcache_dimension *dimension = APR_ARRAY_IDX(tileset->dimensions,i,mapcache_dimension*);
-        if((value = (char*)apr_table_get(argdimensions,dimension->name)) != NULL) {
-          char *tmpval = apr_pstrdup(ctx.pool,value);
-          int ok = dimension->validate(&ctx,dimension,&tmpval);
-          if(GC_HAS_ERROR(&ctx) || ok != MAPCACHE_SUCCESS ) {
-            return usage(argv[0],"failed to validate dimension");
-            return 1;
-          } else {
-            /* validate may have changed the dimension value, so set that value into the dimensions table */
-            apr_table_setn(dimensions,dimension->name,tmpval);
-          }
-        } else {
-          /* a dimension was not specified on the command line, add the default value */
-          apr_table_setn(dimensions, dimension->name, dimension->default_value);
-        }
+    dimensions = apr_array_make(ctx.pool,tileset->dimensions->nelts,sizeof(mapcache_requested_dimension*));
+    for(i=0; i<tileset->dimensions->nelts; i++) {
+      mapcache_dimension *dimension = APR_ARRAY_IDX(tileset->dimensions,i,mapcache_dimension*);
+      mapcache_requested_dimension *rdim = apr_pcalloc(ctx.pool,sizeof(mapcache_requested_dimension));
+      rdim->dimension = dimension;
+      if((value = apr_table_get(argdimensions,dimension->name)) == NULL) {
+        value = dimension->default_value;
       }
+      rdim->requested_value = apr_pstrdup(ctx.pool,value);
+      APR_ARRAY_PUSH(dimensions,mapcache_requested_dimension*)=rdim;
     }
-    if(tileset->timedimension) {
-      if((value = (char*)apr_table_get(argdimensions,tileset->timedimension->key)) != NULL) {
-        apr_array_header_t *timedim_selected = mapcache_timedimension_get_entries_for_value(&ctx,tileset->timedimension, tileset, grid_link->grid, extent, value);
-        if(GC_HAS_ERROR(&ctx) || !timedim_selected) {
-          return usage(argv[0],"failed to validate time dimension");
-        }
-        if(timedim_selected->nelts == 0) {
-          return usage(argv[0],"Time dimension %s=%s returns no configured entry",tileset->timedimension->key,
-                  value);
-        }
-        if(timedim_selected->nelts > 1) {
-          return usage(argv[0],"Time dimension %s=%s returns more than 1 configured entries",tileset->timedimension->key,
-                  value);
-        }
-        apr_table_set(dimensions,tileset->timedimension->key,APR_ARRAY_IDX(timedim_selected,0,char*));
-      } else {
-        return usage(argv[0],"tileset references a TIME dimension, but none supplied on commandline. (hint: -D %s=<timestamp>",tileset->timedimension->key);
-
+    if(tileset->dimension_assembly_type != MAPCACHE_DIMENSION_ASSEMBLY_NONE) {
+      if(!tileset->store_dimension_assemblies) {
+        return usage(argv[0],"cannot seed a layer with dimension assembling and no caching of resulting assembly");
+      }
+      if(tileset->metasize_x != 1 || tileset->metasize_y != 1) {
+        return usage(argv[0],"cannot seed a layer with dimension assembling and metatiling enabled (hint: rerun with -M 1,1 to disable metatiling");
       }
     }
   }
 
-  {
-  /* start the logging thread */
-    //create the queue where the seeding statuses will be put
-    apr_threadattr_t *log_thread_attrs;
-    apr_queue_create(&log_queue,2,ctx.pool);
-
-    //start the rendering threads.
-    apr_threadattr_create(&log_thread_attrs, ctx.pool);
-    apr_thread_create(&log_thread, log_thread_attrs, log_thread_fn, NULL, ctx.pool);
-  }
 
   if(nthreads == 0 && nprocesses == 0) {
     nthreads = 1;
@@ -1273,13 +1409,24 @@ int main(int argc, const char **argv)
   if(nthreads >= 1 && nprocesses >= 1) {
     return usage(argv[0],"cannot set both nthreads and nprocesses");
   }
+  
+  {
+  /* start the logging thread */
+    //create the queue where the seeding statuses will be put
+    apr_threadattr_t *log_thread_attrs;
+    apr_queue_create(&log_queue,MAPCACHE_MAX(nthreads,nprocesses),ctx.pool);
+
+    //start the rendering threads.
+    apr_threadattr_create(&log_thread_attrs, ctx.pool);
+    apr_thread_create(&log_thread, log_thread_attrs, log_thread_fn, NULL, ctx.pool);
+  }
+  
   if(nprocesses > 1) {
 #ifdef USE_FORK
     key_t key;
     int i;
     pid_t *pids = malloc(nprocesses*sizeof(pid_t));
     struct msqid_ds queue_ds;
-    ctx.threadlock = NULL;
     key = ftok(argv[0], 'B');
     if ((msqid = msgget(key, 0644 | IPC_CREAT|S_IRUSR|S_IWUSR)) == -1) {
       return usage(argv[0],"failed to create sysv ipc message queue");
@@ -1314,7 +1461,7 @@ int main(int argc, const char **argv)
         pids[i] = pid;
       }
     }
-    cmd_worker();
+    feed_worker();
     for(i=0; i<nprocesses; i++) {
       int stat_loc;
       waitpid(pids[i],&stat_loc,0);
@@ -1324,26 +1471,48 @@ int main(int argc, const char **argv)
     return usage(argv[0],"bug: multi process support not available");
 #endif
   } else {
-    //start the thread that will populate the queue.
-    apr_thread_mutex_create((apr_thread_mutex_t**)&ctx.threadlock,APR_THREAD_MUTEX_DEFAULT,ctx.pool);
+    apr_threadattr_t *seed_thread_attrs;
     //create the queue where tile requests will be put
     apr_queue_create(&work_queue,nthreads,ctx.pool);
 
+    {
+      /* start the feeding thread */
+      apr_threadattr_t *feed_thread_attrs;
+      //start the rendering threads.
+      apr_threadattr_create(&feed_thread_attrs, ctx.pool);
+      apr_thread_create(&feed_thread, feed_thread_attrs, feed_thread_fn, NULL, ctx.pool);
+    }
+
+
+
     //start the rendering threads.
-    apr_threadattr_create(&thread_attrs, ctx.pool);
-    threads = (apr_thread_t**)apr_pcalloc(ctx.pool, nthreads*sizeof(apr_thread_t*));
+    apr_threadattr_create(&seed_thread_attrs, ctx.pool);
+    seed_threads = (apr_thread_t**)apr_pcalloc(ctx.pool, nthreads*sizeof(apr_thread_t*));
     for(n=0; n<nthreads; n++) {
-      apr_thread_create(&threads[n], thread_attrs, seed_thread, NULL, ctx.pool);
+      if(n && thread_delay > 0) {
+        apr_sleep((int)(thread_delay * 1000000));
+      }
+      apr_thread_create(&seed_threads[n], seed_thread_attrs, seed_thread, NULL, ctx.pool);
     }
-    cmd_worker();
+
+
+    //the worker has finished generating the list of tiles to be seeded, now wait for the rendering threads to finish
     for(n=0; n<nthreads; n++) {
-      apr_thread_join(&rv, threads[n]);
+      apr_thread_join(&rv, seed_threads[n]);
     }
+
+    apr_thread_join(&rv, feed_thread);
   }
   {
+    int retries=0;
+    int ret;
     struct seed_status *st = calloc(1,sizeof(struct seed_status));
     st->status = MAPCACHE_STATUS_FINISHED;
-    apr_queue_push(log_queue,(void*)st);
+    ret = apr_queue_push(log_queue,(void*)st);
+    while (ret == APR_EINTR && retries<10) {
+      retries++;
+      ret = apr_queue_push(log_queue,(void*)st);
+    }
     apr_thread_join(&rv, log_thread);
   }
 
@@ -1351,10 +1520,17 @@ int main(int argc, const char **argv)
     struct mctimeval now_t;
     float duration;
     int ntilestot = n_metatiles_tot*tileset->metasize_x*tileset->metasize_y;
+    int nnodatatot = n_nodata_tot*tileset->metasize_x*tileset->metasize_y;
     mapcache_gettimeofday(&now_t,NULL);
     duration = ((now_t.tv_sec-starttime.tv_sec)*1000000+(now_t.tv_usec-starttime.tv_usec))/1000000.0;
 
-    printf("\nseeded %d metatiles (%d tiles) in %.1f seconds at %.1f tiles/sec\n",n_metatiles_tot, ntilestot, duration, ntilestot/duration);
+    printf("\nseeded %d metatiles (%d total tiles, %d non-empty tiles) in %.1f seconds at %.1f tiles/sec (%.1f non-empty tiles/sec)\n",
+           n_metatiles_tot,
+           ntilestot,
+           ntilestot-nnodatatot,
+           duration,
+           ntilestot/duration,
+           (ntilestot-nnodatatot)/duration);
   } else {
     if(!error_detected) {
       printf("0 tiles needed to be seeded, exiting\n");
